Require Import List.
Require Import Max.
Require Import Omega.


(**
  Notations for lists.
*)
Bind Scope list_scope with list.
Arguments Scope app [type_scope list_scope list_scope].
Arguments Scope cons [type_scope _ list_scope].
Arguments Scope List.In [type_scope _ list_scope].
Infix "::" := cons (at level 60, right associativity) : list_scope.
Infix "++" := app (right associativity, at level 60) : list_scope.
Notation "[ x ; .. ; y ]" := (cons x .. (cons y nil) ..) : list_scope.


(**
  List membership theorems.
*)
Lemma list_not_in_cons_l :
  forall (A : Type) (x : A) y M, {x = y} + {x <> y} -> ~ In x (y :: M) -> x <> y.
Proof.
  intros A x; induction M as [ | m ms IH ]; simpl.
  intros dec H.
    case dec; try subst x; auto.
  intros dec H.
    case dec; try subst x; auto.
Qed.


(**
  Facts about arithmetic.
*)

Lemma max_lt_l :
  forall (x y z : nat), x <= y -> x <= max y z.
Proof.
  induction x; auto with arith.
  induction y; induction z; simpl; auto with arith.
Qed.

Hint Resolve max_lt_l : arith.


Lemma max_lt_r :
  forall (x y z : nat), x <= z -> x <= max y z.
Proof.
  intros x y z H.
  rewrite max_comm; apply max_lt_l; trivial.
Qed.

Hint Resolve max_lt_r : arith.


Lemma le_lt_S :
  forall (n m : nat),
  n <= m -> n < S m.
Proof.
  intros; auto with arith.
Qed.

Hint Resolve le_lt_S : arith.


Lemma lt_Sn_n :
  forall (n m : nat),
  n < S m -> n <> m -> n < m.
Proof.
  intros; omega.
Qed.

Hint Resolve lt_Sn_n : arith.


Lemma finite_nat_list_max :
  forall (l : list nat),
  { n : nat | forall x, In x l -> x <= n }.
Proof.
  induction l as [ | l ls IHl ].
  exists 0; intros x H; inversion H.
  inversion IHl as [x H].
  exists (max x l); intros y J; simpl in J; inversion J.
    subst; auto with arith.
    assert (y <= x); auto with arith.
Qed.


Lemma finite_nat_list_max' :
  forall (l : list nat),
  { n : nat | ~ In n l }.
Proof.
  intros l.
  case (finite_nat_list_max l); intros x H.
  exists (S x).
  intros J.
  assert (K := H _ J); omega.
Qed.
